#!/usr/bin/perl
use strict;

unless ($#ARGV == 1){
  print STDERR 'Usage: perl paraphrase_scorer.pl $prediction_file $gold_file';
  print STDERR "\n";
  exit;
}

my ($prediction_file,$gold_file) = @ARGV;

open PRED, $prediction_file or die "Error reading $prediction_file\n$!";

my $errors = 0;
my %ranks;
my %lines;
my %paraphrases;
my %scores;

while (my $line = <PRED>){
    chomp $line;
    if ($line =~ /^(\d+)\t+([a-z -]+)\t+([a-z -]+)(?:\t+([0-9.]+))?\s*$/){
	unless (defined $ranks{$2}){
	    $ranks{$2} = [];
	    $lines{$2} = [];
	    $paraphrases{$2} = [];
	    $scores{$2} = [];
	}
	push @{$ranks{$2}}, $1;
	push @{$lines{$2}}, $line;
	push @{$paraphrases{$2}}, $3;
	push @{$scores{$2}}, $4 if $4;
    }else{
	print "Error at line: $line\n";
	$errors++;
    }
}

close PRED;
my $has_scores = 0;
my $no_scores = 0;

for my $c (keys %ranks){
    if ($#{$scores{$c}} == -1){
	$no_scores++;
    }elsif ($#{$scores{$c}} != $#{$ranks{$c}}){
	$errors++;
	$has_scores++;
	print "Not all paraphrases for $c have scores; skipping $c \n";
	next;
    }else{
	$has_scores++;
    }
	
    my @sorted = sort {$a <=> $b} @{$ranks{$c}};
    my @sorted_lines = @{$lines{$c}}[sort {$ranks{$c}[$a] <=> $ranks{$c}[$b]} 0 .. $#{$lines{$c}}];
    my $expected = 1;
    my $i;
    while ($i <= $#sorted){
	if ($sorted[$i] != $expected){
	    print "Expected to see rank $expected at line: $sorted_lines[$i]\n";
	    $errors++;
	    $expected = $sorted[$i]+1;
	}else{
	    while ($i < $#sorted && $sorted[$i] == $sorted[$i+1]){
		$expected++;
		$i++;
	    }
	}
	$i++;
	$expected++;
    }
}

if ($errors){
    print STDERR "Error: $errors lines had errors\n";
    exit;
}elsif ($has_scores and $no_scores){
    print STDERR "Error: If scores are provided for some compounds they must be provided for all compounds\n";
    exit;
}else{
    print "Format is correct\n";
}

my %pred_ranks;
my %pred_scores;

for my $c (keys %ranks){
    for my $i (0 .. $#{$ranks{$c}}){
	$pred_ranks{$c}{$paraphrases{$c}[$i]} = $ranks{$c}[$i];
	$pred_scores{$c}{$paraphrases{$c}[$i]} = $scores{$c}[$i] if $has_scores;
    }
}

my %gold;

open GOLD, $gold_file or die "Error reading $gold_file\n$!";
while (<GOLD>){
    chomp;
    unless (/^([a-z -]+)\t+([a-z -]+)\t+(\d+)\s*$/){
      print STDERR "Error: Badly-formatted gold standard: $_\n";
      exit;
    }
    $gold{$1}{$2} = $3;
}
close GOLD;

for my $c (keys %pred_ranks){
    for my $p (keys %{$pred_ranks{$c}}){
      unless (defined $gold{$c}{$p}){
	print STDERR "Error: Paraphrase \"$p\" for compound \"$c\" does not appear in the gold standard\n";
	exit;
      }
    }
}

my $spearman_ave;
my $kl_ave;
my $pearson_ave;

print "\nPer-compound scores:\n";

for my $c (keys %gold){
    my @g;
    my @pr;
    my @ps;
    for my $p (keys %{$gold{$c}}){
      unless(defined $pred_ranks{$c}{$p}){
	print STDERR "Error: Paraphrase \"$p\" for compound \"$c\" does not appear in the prediction file\n";
	exit;
      }
	push @g, $gold{$c}{$p};
	push @pr, $pred_ranks{$c}{$p};
	push @ps, $pred_scores{$c}{$p} if $has_scores;
    }

    my $rho = -1*spearman(\@g,\@pr); # multiply by -1 because high scores => low ranks
    my $r = corr(\@g,\@ps) if $has_scores;
    my $kl = kullback(\@g,\@ps) if $has_scores;
    print "$c: Spearman correlation $rho";
    print ", Pearson correlation $r, Kullback-Leibler divergence $kl" if $has_scores;
    print "\n";
    $spearman_ave += $rho;
    $pearson_ave += $r;
    $kl_ave += $kl;
} 

$spearman_ave /= scalar keys %gold;
$pearson_ave /= scalar keys %gold;
$kl_ave /= scalar keys %gold;

print "\nAverage Spearman correlation: $spearman_ave (Official evaluation measure)\n";
if ($has_scores){
    print "Average Pearson correlation: $pearson_ave\nAverage Kullback-Leibler divergence: $kl_ave\n";
}

sub kullback{
    my @x = @{$_[0]}; # the gold distribution
    my @y = @{$_[1]}; # the predicted distribution

    # normalise the distributions
    my $x_norm = 0;
    my $y_norm = 0;
    for my $i (0 .. $#x){
	$x_norm += $x[$i];
	$y_norm += $y[$i];
    }

    for my $i (0 .. $#x){
	$x[$i] /= $x_norm;
	$y[$i] /= $y_norm;
    }

    # calculate the divergence
    my $kl = 0;
    for my $i (0 .. $#x){
	$kl += $x[$i]*log($x[$i]/$y[$i]);
    }

    return $kl/log(2);
}

sub spearman{
    my @x = @{$_[0]};
    my @y = @{$_[1]};

    my @x_rank;
    my @y_rank;

    my %x_scores;
    for my $i (0 .. $#x){
	if (defined $x_scores{$x[$i]}){
	    push @{$x_scores{$x[$i]}}, $i;
	}else{
	    $x_scores{$x[$i]} = [$i];
	}
    }

    my $c = 1;
    for my $i (sort {$a <=> $b} keys %x_scores){
	my $r;
	for my $s (@{$x_scores{$i}}){
	    $r += $c;
	    $c++;
	}
	$r /= scalar @{$x_scores{$i}};
	for my $s (@{$x_scores{$i}}){
	    $x_rank[$s] = $r;
	}
    }

    my %y_scores;
    for my $i (0 .. $#y){
	if (defined $y_scores{$y[$i]}){
	    push @{$y_scores{$y[$i]}}, $i;
	}else{
	    $y_scores{$y[$i]} = [$i];
	}
    }
    $c = 1;
    for my $i (sort {$a <=> $b} keys %y_scores){
	my $r;
	for my $s (@{$y_scores{$i}}){
	    $r += $c;
	    $c++;
	}
	$r /= scalar @{$y_scores{$i}};
	for my $s (@{$y_scores{$i}}){
	    $y_rank[$s] = $r;
	}
    }

    return corr(\@x_rank,\@y_rank);
}

sub corr{
    my @x = @{$_[0]};
    my @y = @{$_[1]};

    my $mean_x;
    my $mean_y;
    my $n_x = $#x+1;
    my $n_y = $#y + 1;

    for my $x(@x){
	$mean_x += $x;
    }
    $mean_x /= $n_x;

    for my $y (@y){
	$mean_y += $y;
    }
    $mean_y /= $n_y;

    my $cov = 0;
    my $sd_x = 0;
    my $sd_y = 0;

    for my $i (0 .. $#x){
	$cov += ($x[$i] - $mean_x) * ($y[$i] - $mean_y);
	$sd_x += ($x[$i] - $mean_x)*($x[$i] - $mean_x);
	$sd_y += ($y[$i] - $mean_y) * ($y[$i] - $mean_y);
    }

    return $cov/(sqrt($sd_x)*sqrt($sd_y));
}
